from typing import Type, Tuple
from typing_inspect import get_args, is_optional_type, is_generic_type, get_origin  # type: ignore


__all__ = ["optional_core", "generic_core", "is_optional_type", "is_generic_type", "get_origin"]


def optional_core(t: Type) -> Tuple[bool, Type]:
    """
    :return  (is_optional, t.core | t | NoneType)
    """
    if t is type(None):
        return True, type(None)
    if is_optional_type(t):
        first, second = get_args(t)
        return True, (second if isinstance(None, first) else first)
    else:
        return False, t


def generic_core(t: Type):
    args = get_args(t)
    return args[0]



from typing import Type, TypeVar, Iterable, List, Iterator, Generic, Optional, Union, Dict, Any
from enum import Enum
from contextlib import contextmanager

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.orm.session import Session

from ..context import Context
from ..application import Application
from ..storage import Storage
from ..typehint import optional_core


__all__ = ["DbPlugin", "DbServ", "Mapper"]


class DatabaseKey(Enum):
    session = 1


class DbPlugin:
    patterns: Iterable[str]
    db_session_maker: Any
    db_engine: Any
    autocommit: bool

    def __init__(self,
                 uri: str,
                 echo: bool=True,
                 autoflush: bool=True,
                 autocommit: bool=False,
                 patterns: Iterable[str]=('.*',)):
        self.patterns = patterns
        if '://' not in uri:
            uri = 'sqlite:///' + uri
        engine = create_engine(uri, pool_recycle=3600)
        engine.echo = echo
        self.db_session_maker = scoped_session(sessionmaker(
                autoflush=autoflush, autocommit=autocommit, bind=engine))
        self.db_engine = engine
        self.autocommit = autocommit

    def processor(self, ctx: Context):
        db = self.db_session_maker()
        try:
            ctx.box[DatabaseKey.session] = db
            if self.autocommit:
                with db.begin():
                    return ctx()
            else:
                return ctx()
        except:
            db.rollback()
            raise
        finally:
            db.close()

    def init_app(self, app: Application):
        for pattern in self.patterns:
            segs = pattern.split()
            if len(segs) == 1:
                app.add_interceptor(pattern, method='*', dealer=self.processor)
            elif len(segs) == 2:
                app.add_interceptor(segs[1], method=segs[0], dealer=self.processor)

    def teardown(self, exception: Exception):
        pass

    @contextmanager
    def make_session(self) -> Iterator[Session]:
        session: Session = self.db_session_maker()
        try:
            yield session
        except Exception as e:
            session.rollback()
            raise e
        finally:
            session.close()


T = TypeVar('T')


class DbServ:
    ctx: Context

    @property
    def db(self) -> Session:
        session = self.ctx.box.get(DatabaseKey.session)
        if session is None:
            raise ValueError('database session not available')
        return session

    def mapper(self, cls: Type[T]) -> 'Mapper[T]':
        return Mapper(self.db, cls)


def table(*, name: str):
    def g(cls):
        setattr(cls, '__tablename__', name)
        return cls
    return g


def transient(*props: str):
    def g(cls):
        if props:
            cls_props = Storage.type_hints(cls)
            for prop in props:
                if prop not in cls_props:
                    raise AttributeError(f"class {cls} has no attribute '{prop}' for @transient")
            setattr(cls, '__transient__', props)
        return cls
    return g


class Mapper(Generic[T]):
    session: Session
    model_type: Type[T]
    tablename: str
    model_schema: Storage
    primary_key: str
    _where_sqls: List[str]
    _where_data: Dict
    _orderby_sql: str = ''

    def __init__(self, session: Session, cls: Type[T]) -> None:
        self.session = session
        self.model_type = cls
        self.tablename = getattr(cls, '__tablename__', '') or cls.__name__
        self.model_schema = Storage.type_hints(cls)
        if hasattr(cls, '__transient__'):
            self.model_schema -= getattr(cls, '__transient__')
        if not self.model_schema:
            raise TypeError('%s schema is empty!' % cls)
        self.primary_key = next(iter(self.model_schema))
        self._where_sqls = []
        self._where_data = {}

    def _where_clause(self) -> str:
        return ' and '.join(f'({s})' for s in self._where_sqls)

    def _full_select_sql(self) -> str:
        where_clause = self._where_clause()
        titles = ','.join(f'`{key}`' for key in self.model_schema)
        if where_clause:
            sql = f'SELECT {titles} FROM `{self.tablename}` WHERE {where_clause}'
        else:
            sql = f'SELECT {titles} FROM `{self.tablename}`'
        if self._orderby_sql:
            sql += f' ORDER BY {self._orderby_sql}'
        return sql

    def _imtransient_storage(self, obj: T) -> Storage:
        obj_storage = Storage.of(obj)
        if hasattr(self, '__transient__'):
            return obj_storage - getattr(self, '__transient__')
        else:
            return obj_storage

    def bridge(self, row) -> T:
        obj = self.model_type()
        for key, val in row.items():
            if key in self.model_schema:
                prop_type = self.model_schema[key]
                is_optional, prop_type = optional_core(prop_type)
                if val is not None:
                    if isinstance(val, prop_type):
                        setattr(obj, key, val)
                    else:
                        setattr(obj, key, prop_type(val))
                elif val is None and is_optional:
                    setattr(obj, key, None)
        return obj

    def select_count(self) -> int:
        where_clause = self._where_clause()
        if where_clause:
            sql = f'SELECT COUNT(1) FROM `{self.tablename}` WHERE {where_clause}'
        else:
            sql = f'SELECT COUNT(1) FROM `{self.tablename}`'
        return self.session.execute(sql, self._where_data).scalar()

    def select_first(self) -> Optional[T]:
        sql = self._full_select_sql() + ' LIMIT 1'
        row = self.session.execute(sql, self._where_data).first()
        if row is None:
            return None
        return self.bridge(row)

    def select(self) -> List[T]:
        ret = []
        sql = self._full_select_sql()
        rows = self.session.execute(sql, self._where_data)
        for row in rows:
            ret.append(self.bridge(row))
        return ret

    def insert(self, obj: T, commit: bool=True) -> None:
        obj_storage = self._imtransient_storage(obj)
        titles = ','.join(f'`{key}`' for key in obj_storage.keys())
        slots = ','.join(f':{key}' for key in obj_storage.keys())
        sql = f'INSERT INTO `{self.tablename}` ({titles}) VALUES ({slots})'
        result = self.session.execute(sql, obj_storage)
        setattr(obj, self.primary_key, result.lastrowid)
        if commit:
            self.session.commit()

    def insert_if_not_exist(self, obj: T, commit: bool=True) -> None:
        obj_storage = self._imtransient_storage(obj)
        titles = ','.join(f'`{key}`' for key in obj_storage.keys())
        slots = ','.join(f':{key}_1' for key in obj_storage.keys())
        data = {f'{key}_1': val for key, val in obj_storage.items()}
        data.update(self._where_data)
        where_clause = self._where_clause()
        if not where_clause:
            raise ValueError(f'WHERE clause cannot be empty for INSERT_IF_NOT_EXIST!')
        sql = f'INSERT INTO `{self.tablename}` ({titles}) SELECT {slots}' + \
            f' WHERE not exists ( SELECT 1 FROM `{self.tablename}` WHERE {where_clause})'
        self.session.execute(sql, data)
        if commit:
            self.session.commit()

    def update(self, obj: T, commit: bool=True) -> None:
        obj_storage = self._imtransient_storage(obj)
        set_sql = ', '.join(f'`{key}`=:{key}_1' for key in obj_storage.keys())
        data = {f'{key}_1': val for key, val in obj_storage.items()}
        data.update(self._where_data)
        where_clause = self._where_clause()
        if not where_clause:
            raise ValueError(f'WHERE clause cannot be empty for UPDATE!')
        sql = f'UPDATE `{self.tablename}` SET {set_sql} WHERE {where_clause}'
        self.session.execute(sql, data)
        if commit:
            self.session.commit()

    def increment(self, obj: T, commit: bool=True) -> None:
        obj_storage = self._imtransient_storage(obj)
        obj_dict = {key: val for key, val in obj_storage.items() if isinstance(val, int)}
        set_sql = ', '.join(f'`{key}`=`{key}`+(:{key}_1)' for key in obj_dict.keys())
        data = {f'{key}_1': val for key, val in obj_dict.items()}
        data.update(self._where_data)
        where_clause = self._where_clause()
        if not where_clause:
            raise ValueError(f'WHERE clause cannot be empty for increment UPDATE!')
        sql = f'UPDATE `{self.tablename}` SET {set_sql} WHERE {where_clause}'
        self.session.execute(sql, data)
        if commit:
            self.session.commit()

    def delete(self, commit: bool=True) -> None:
        where_clause = self._where_clause()
        if not where_clause:
            raise ValueError(f'WHERE clause cannot be empty for DELETE!')
        sql = f'DELETE FROM `{self.tablename}` WHERE {where_clause}'
        self.session.execute(sql, self._where_data)
        if commit:
            self.session.commit()

    def by_id(self, primary_key: Union[int, str]) -> 'Mapper[T]':
        sql = f'`{self.primary_key}`=:{self.primary_key}'
        data = {self.primary_key: primary_key}
        self._where_sqls.append(sql)
        self._where_data.update(data)
        return self

    def and_equal(self, obj: Union[T, Dict[str, Any]]) -> 'Mapper[T]':
        data = obj if isinstance(obj, dict) else self._imtransient_storage(obj)
        self._where_data.update(data)
        for key, val in data.items():
            sql = f'`{key}`=:{key}'
            self._where_sqls.append(sql)
        return self

    def and_(self, clause: str, data: Storage) -> 'Mapper[T]':
        self._where_sqls.append(clause)
        self._where_data.update(data)
        return self

    def order_desc(self) -> 'Mapper[T]':
        self._orderby_sql = f'`{self.primary_key}` desc'
        return self

    def order_by(self, clause: str) -> 'Mapper[T]':
        self._orderby_sql = clause
        return self



from pathlib import Path
import yaml
import json


def glob_find(ref_path):
    yield from Path(ref_path).glob('*.json')
    yield from Path(ref_path).glob('*.yaml')


def dir_mtime(ref_path):
    return max(p.stat().st_mtime for p in glob_find(ref_path))


def packed(ref_path, title, version):
    """
    ref_path -> packed_swagger_data
    """
    dup_path_keys = set()
    dup_component_keys = set()
    swagger_data = {}
    for p in glob_find(ref_path):
        if p.suffix == '.yaml':
            cur_api = yaml.full_load(p.open('r', encoding='utf-8'))
        else:
            cur_api = json.load(p.open('r', encoding='utf-8'))
        if not swagger_data:
            swagger_data.update(cur_api)
            swagger_data['paths'] = {}
            swagger_data['components'] = {'schemas': {}}
        for key, val in cur_api.get('paths', {}).items():
            if key in swagger_data['paths']:
                dup_path_keys.add(key)
            else:
                swagger_data['paths'][key] = val

        for key, val in cur_api.get('components', {}).get('schemas', {}).items():
            if key in swagger_data['components']['schemas']:
                dup_component_keys.add(key)
            else:
                swagger_data['components']['schemas'][key] = val
    if not swagger_data:
        print('Empty reference directory!')
        exit(1)

    if dup_path_keys or dup_component_keys:
        if dup_path_keys:
            print('Duplicated paths keys:')
            for key in dup_path_keys:
                print('    ' + key)
        if dup_component_keys:
            print('Duplicated components keys:')
            for key in dup_component_keys:
                print('    ' + key)
        print('\nLoad swagger failed!')
        exit(1)

    swagger_data['info']['title'] = title
    swagger_data['info']['version'] = version
    return swagger_data


# def unpack(json_path):
#     """
#     json_path_file -> parts/(api.json|paths|components)
#     """
#     SwaggerData.init(json_path=json_path)
#     Path('parts').mkdir(parents=True, exist_ok=True)
#     Path('parts/paths').mkdir(parents=True, exist_ok=True)
#     Path('parts/components').mkdir(parents=True, exist_ok=True)
#     swagger_data = SwaggerData.get()
#     for filename in sorted(os.listdir('parts/paths')):
#         os.remove(f'parts/paths/{filename}')
#     if 'paths' in swagger_data:
#         for path_key, path_val in swagger_data['paths'].items():
#             filename = path_key.replace('/', '_')
#             content = json.dumps({'key': path_key, 'value': path_val}, ensure_ascii=False, indent=2)
#             with Path(f'parts/paths/{filename}.json').open('w', encoding='utf-8') as f:
#                 f.write(content)
#
#     for filename in sorted(os.listdir('parts/components')):
#         os.remove(f'parts/components/{filename}')
#     if 'components' in swagger_data and 'schemas' in swagger_data['components']:
#         for comp_key, comp_val in swagger_data['components']['schemas'].items():
#             content = json.dumps({'key': comp_key, 'value': comp_val}, ensure_ascii=False, indent=2)
#             with Path(f'parts/components/{comp_key}.json').open('w', encoding='utf-8') as f:
#                 f.write(content)
#
#     swagger_data['paths'] = {}
#     swagger_data['components'] = {}
#     content = json.dumps(swagger_data, ensure_ascii=False, indent=2)
#     with Path(f'parts/api.json').open('w', encoding='utf-8') as f:
#         f.write(content)
#
#
# def pack(json_path):
#     """
#     parts/* -> json_path_file
#     """
#     swagger_data = json.load(open('parts/api.json'))
#     swagger_data.setdefault('paths', {})
#     swagger_data.setdefault('components', {})
#     swagger_data['components'].setdefault('schemas', {})
#     for filename in sorted(os.listdir('parts/paths')):
#         if filename.endswith('.json'):
#             content = json.load(open(f'parts/paths/{filename}'))
#             key, value = content['key'], content['value']
#             swagger_data['paths'][key] = value
#
#     for filename in sorted(os.listdir('parts/components')):
#         if filename.endswith('.json'):
#             content = json.load(open(f'parts/components/{filename}'))
#             key, value = content['key'], content['value']
#             swagger_data['components']['schemas'][key] = value
#
#     with open(json_path, 'w', encoding='utf-8') as fp:
#         json.dump(swagger_data, fp=fp, ensure_ascii=False, indent=2)



# 存放与类型转换有关的类型定义，且不依赖同级其他库
from enum import Enum
from datetime import datetime as Datetime
from json import JSONEncoder
from itertools import chain
from typing import Type, List, Callable, Union, Dict, Any
from .storage import Storage


__all__ = ["uint", "Jsonizable", "ParamStr", "MultipartFile", "JsonBridgeFunc"]


class uint(int):
    def __init__(self, v):
        super().__init__()
        if self < 0:
            raise ValueError("invalid range for uint(): '%d'" % self)


Jsonizable = Union[str, int, float, Dict, List, None]


class ParamStr(str):
    pass


class MultipartFile:
    filename: str
    value: bytes

    def __init__(self, upfile):
        self.filename = upfile.filename
        self.value = upfile.value

    def __str__(self) -> str:
        return f'<MultipartFile filename={self.filename} value={str(self.value)}>'


JsonBridgeFunc = Callable[[Any], Jsonizable]


def default_response_bridge(obj: Any) -> Jsonizable:
    if isinstance(obj, Datetime):
        return obj.isoformat()
    if isinstance(obj, Enum):
        return obj.value
    if isinstance(obj, bytes):
        return str(obj)
    return None


def make_response_encoder(bridge_funcs: List[JsonBridgeFunc]):
    class ResponseEncoder(JSONEncoder):
        def default(self, obj):
            if obj is None:
                return obj
            for bridge_func in chain(bridge_funcs, [default_response_bridge]):
                dest_val = bridge_func(obj)
                if dest_val is not None:
                    return dest_val
            try:
                if Storage.type_hints(type(obj)):
                    return Storage.of(obj)
            except:
                pass

    return ResponseEncoder



from typing import Any
import sys
NEW_INSPECT = sys.version_info[:3] >= (3, 8, 0)
if NEW_INSPECT:
    from typing import Protocol  # since python3.8+
else:
    from typing_extensions import Protocol  # type: ignore


__all__ = ["PluginProto"]


class PluginProto(Protocol):
    def init_app(self, app: Any) -> None:
        ...

    def teardown(self, exception: Exception) -> None:
        ...



from typing import Dict, List, Tuple
import json


def deep_copy(obj):
    return json.loads(json.dumps(obj))


def make_row(level, name, types, description, additional, ref=''):
    return {'level': level, 'name': name, 'types': types, 'description': description, 'additional': additional, 'ref': ref}


def jsonschema_from_ref(ref: str, swagger_data: Dict) -> Dict:
    ret = swagger_data
    if '#' in ref:
        ref = ref.rsplit('#', 1)[-1]
    for seg in ref.split('/'):
        if not seg: continue
        ret = ret.get(seg, {})
    return ret


def jsonschema_to_rows(parent_level: str, name: str, schema: Dict, swagger_data: Dict, *, required: List) -> List:
    level = f'{parent_level}{name}/'
    additional = deep_copy(schema)
    for key in ['description', '$ref', 'type', 'items', 'properties']:
        additional.pop(key, '')
    description = schema.get('description', '')
    if '$ref' in schema:
        ref_schema = jsonschema_from_ref(schema['$ref'], swagger_data)
        ret = jsonschema_to_rows(parent_level, name, ref_schema, swagger_data, required=[])
        if description:
            ret[0]['description'] = description
        if additional:
            ret[0]['additional'] = additional
        return ret
    types = schema.get('type', [])
    if not isinstance(types, list):
        types = [types]
    if name in required:
        types.append('!')
    if 'array' in types:
        items = schema.get('items', {})
        ret = [make_row(level, name, types, description, additional)]
        if items:
            ret.extend(jsonschema_to_rows(level, '[ ]', items, swagger_data, required=[]))
        return ret
    elif 'object' in types:
        properties = schema.get('properties', {})
        ret = [make_row(level, name, types, description, additional)]
        for key, val in properties.items():
            ret.extend(jsonschema_to_rows(level, key, val, swagger_data, required=additional.get('required', [])))
        return ret
    else:
        return [make_row(level, name, types, description, additional)]


def compare_json(level, actual, expect, full_actual, full_expect) -> List[str]:
    def item_to_pair(item) -> Tuple:
        if isinstance(item, str):
            return item, item
        if 'name' in item and 'in' in item:
            return item['name'], item
        if 'level' in item and 'types' in item:
            return item['level'], item
        if 'rows' in item and 'content_type' in item:
            return item.get('status_code', '') + ':' + item['content_type'], item['rows']
        raise NotImplementedError(item)

    if type(actual) != type(expect):
        if actual or expect:
            return [f'{level} 类型不匹配 actual:{repr(actual)} expect:{repr(expect)}']
        else:
            return []
    if isinstance(actual, dict):
        ret = []
        if '$ref' in actual:
            actual = jsonschema_from_ref(actual['$ref'], full_actual)
        if '$ref' in expect:
            expect = jsonschema_from_ref(expect['$ref'], full_expect)

        actual_keys = actual.keys() - {'format'} if actual.get('type') == 'integer' and actual.get('format') == 'int32' else actual.keys()
        expect_keys = expect.keys() - {'format'} if expect.get('type') == 'integer' and expect.get('format') == 'int32' else expect.keys()
        if 'description' not in actual_keys or 'title' not in actual_keys:
            expect_keys -= {'description', 'title'}

        for key in actual_keys | expect_keys:
            ret.extend(compare_json(f'{level}:{key}', actual.get(key), expect.get(key), full_actual, full_expect))
        return ret
    elif isinstance(actual, list):
        actual_dict = dict(item_to_pair(item) for item in actual)
        expect_dict = dict(item_to_pair(item) for item in expect)
        return compare_json(level, actual_dict, expect_dict, full_actual, full_expect)
    elif actual != expect:
        return [f'{level} 不一致 actual:{repr(actual)} expect:{repr(expect)}']
    else:
        return []
